/*                                                                           
**  fi_lib++  --- A fast interval library (Version 2.0)                     
**                                                                  
**  Copyright (C) 2001:                                                        
**                                                     
**  Werner Hofschuster, Walter Kraemer                               
**  Wissenschaftliches Rechnen/Softwaretechnologie (WRSWT)  
**  Universitaet Wuppertal, Germany                                           
**  Michael Lerch, German Tischler, Juergen Wolff von Gudenberg       
**  Institut fuer Informatik                                         
**  Universitaet Wuerzburg, Germany                                           
** 
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/
#include <stdio.h>
#include <sys/types.h>

#if defined(__i386__)
#include <stdlib.h>
#include <time.h>

#if defined(__linux) && defined(__i386__)
typedef unsigned short uint16_t;
#endif

/* uint16_t const precision_single_bit			= 0x0; */
uint16_t const precision_double_bit			= 0x2;
/* uint16_t const precision_extended_bit			= 0x3; */

/* uint16_t const precision_mask_single			= precision_single_bit << 8; */
uint16_t const precision_mask_double			= precision_double_bit << 8;
/* uint16_t const precision_mask_extended			= precision_extended_bit << 8; */

uint16_t const rounding_tonearest_bit			= 0x0;
uint16_t const rounding_down_bit			= 0x1;
uint16_t const rounding_up_bit				= 0x2;
/* uint16_t const rounding_tozero_bit			= 0x3; */

uint16_t const rounding_mask_tonearest			= rounding_tonearest_bit << 10; 
uint16_t const rounding_mask_down			= rounding_down_bit << 10; 
uint16_t const rounding_mask_up				= rounding_up_bit << 10; 
/* uint16_t const rounding_mask_tozero			= rounding_tozero_bit << 10; */

uint16_t const exception_mask_invalid_operation		= 0x1 << 0;
uint16_t const exception_mask_denormalized_operand	= 0x1 << 1;
uint16_t const exception_mask_zero_divide		= 0x1 << 2;
uint16_t const exception_mask_overflow			= 0x1 << 3;
uint16_t const exception_mask_underflow			= 0x1 << 4;
uint16_t const exception_mask_precision			= 0x1 << 5;

uint16_t const exception_mask_all 			=
	exception_mask_invalid_operation	|
	exception_mask_denormalized_operand	|
	exception_mask_zero_divide		|
	exception_mask_overflow			|
	exception_mask_underflow		|
	exception_mask_precision;

uint16_t const cw_tonearest = 
	exception_mask_all | rounding_mask_tonearest | precision_mask_double;
uint16_t const cw_down = 
	exception_mask_all | rounding_mask_down | precision_mask_double;
uint16_t const cw_up = 
	exception_mask_all | rounding_mask_up | precision_mask_double;
/* uint16_t const cw_tozero = 
	exception_mask_all | rounding_mask_tozero | precision_mask_double; */

#define OPS 200*1000*1000

void add_switched(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"label_as:\n "
 		"fldcw %4; fldl %0 ; fldl %1 ; faddp ; fstpl %3; "
		"fldcw %5; fldl %0 ; fldl %1 ; faddp ; fstpl %3; "
		"decl %2 ; jne label_as; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void add_switchedback(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"label_asb:\n"
		"fldcw %4; fldl %0 ; fldl %1 ; faddp ; fstpl %3; "
		"fldcw %5; fldl %0 ; fldl %1 ; faddp ; fstpl %3; fldcw %6 ; "
		"decl %2 ; jne label_asb; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up), "m" (cw_tonearest) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void add_onesided_switched(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"label_os:\n"
		"fldcw %4; fldl %0 ; fldl %1 ; faddp ; fstpl %3; "
		"fldl %0 ; fchs ; fldl %1 ; fchs ; faddp ; fchs ; fstpl %3; "
		"decl %2 ; jne label_os; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void add_onesided_switchedback(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"label_osb:\n"
		"fldcw %4; fldl %0 ; fldl %1 ; faddp ; fstpl %3; "
		"fldl %0 ; fchs ; fldl %1 ; fchs ; faddp ; fchs ; fstpl %3; fldcw %5; "
		"decl %2 ; jne label_osb; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void add_onesided_global(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"fldcw %4 ; \n"
		"label_og:\n"
		"fldl %0 ; fldl %1 ; faddp ; fstpl %3; "
		"fldl %0 ; fchs ; fldl %1 ; fchs ; faddp ; fchs ; fstpl %3; "
		"decl %2 ; jne label_og; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void sub_switched(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"sub_label_as:\n"
		"fldcw %4; fldl %0 ; fldl %1 ; fsubp ; fstpl %3; "
		"fldcw %5; fldl %0 ; fldl %1 ; fsubp ; fstpl %3; "
		"decl %2 ; jne sub_label_as; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void sub_switchedback(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"sub_label_asb:"
		"fldcw %4; fldl %0 ; fldl %1 ; fsubp ; fstpl %3; "
		"fldcw %5; fldl %0 ; fldl %1 ; fsubp ; fstpl %3; fldcw %6 ; "
		"decl %2 ; jne sub_label_asb; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up), "m" (cw_tonearest) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void sub_onesided_switched(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"sub_label_os:\n"
		"fldcw %4; fldl %0 ; fldl %1 ; fsubp ; fstpl %3; "
		"fldl %0 ; fchs ; fldl %1 ; fchs ; fsubp ; fchs ; fstpl %3; "
		"decl %2 ; jne sub_label_os; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void sub_onesided_switchedback(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"sub_label_osb:\n"
		"fldcw %4; fldl %0 ; fldl %1 ; fsubp ; fstpl %3; "
		"fldl %0 ; fchs ; fldl %1 ; fchs ; fsubp ; fchs ; fstpl %3; fldcw %5; "
		"decl %2 ; jne sub_label_osb; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

void sub_onesided_global(double const &a, double const &b, double & c)
{
	unsigned int count = OPS;
	
	asm 
	(
   		"fldcw %4 ; \n"
		"sub_label_og:\n"
		"fldl %0 ; fldl %1 ; fsubp ; fstpl %3; "
		"fldl %0 ; fchs ; fldl %1 ; fchs ; fsubp ; fchs ; fstpl %3; "
		"decl %2 ; jne sub_label_og; fldl %1" :
		: "m" (a), "m" (b), "r" (count), "m" (c), "m" (cw_down), "m" (cw_up) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)"
	);
}

int main()
{
	double a = 2.0, b = 3.0, c;
	clock_t before, after;

	before = clock();
	add_switched(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native switched add: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		(double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC)/1e6
	);

	before = clock();
	add_switchedback(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native switchedback add: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		(double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC)/1e6
	);

	before = clock();
	add_onesided_switched(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native onesided_switched add: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		(double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC)/1e6
	);

	before = clock();
	add_onesided_switchedback(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native onesided_switchedback add: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		(double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC)/1e6
	);

	before = clock();
	add_onesided_global(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native onesided_global add: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		(double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC)/1e6
	);

	before = clock();
	sub_switched(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native switched sub: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		((double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC))/1e6
	);

	before = clock();
	sub_switchedback(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native switchedback sub: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		((double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC))/1e6
	);

	before = clock();
	sub_onesided_switched(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native onesided_switched sub: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		((double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC))/1e6
	);

	before = clock();
	sub_onesided_switchedback(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native onesided_switchedback sub: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		((double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC))/1e6
	);

	before = clock();
	sub_onesided_global(a,b,c);
	after = clock();

	printf("Values: %f %f %f, Expected: %f %f %f\n", a,b,c, 2.0, 3.0, 5.0);
	printf("Native onesided_global sub: Clocks: %u, Seconds: %f Mops: %f\n", 
		(after-before), 
		(double)(after-before)/(double)CLOCKS_PER_SEC,
		((double)(OPS)/((double)(after-before)/(double)CLOCKS_PER_SEC))/1e6
	);
}
#else
int main()
{}
#endif

