#if ! defined(FILIB_AUTODIFF)

#include <iostream>
#include <cmath>

namespace autodiff
{
	template<typename N, size_t k> struct autodiff;

	template<typename N,size_t k> autodiff<N,k> operator+(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		);
	template<typename N,size_t k> autodiff<N,k> operator-(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		);
	template<typename N,size_t k> autodiff<N,k> operator*(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		);
	template<typename N,size_t k> autodiff<N,k> operator/(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		);
	template<typename N,size_t k> std::ostream & operator<<(
		std::ostream & out,
		autodiff<N,k> const & obj);

	template<typename N,size_t k>
	autodiff<N,k> sin(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> cos(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> tan(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> cot(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> asin(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> acos(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> atan(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> acot(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> sinh(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> cosh(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> tanh(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> coth(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> asinh(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> acosh(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> atanh(autodiff<N,k> const & A);
	template<typename N,size_t k>
	autodiff<N,k> acoth(autodiff<N,k> const & A);

	template <typename N = double, size_t k = 1> struct autodiff
	{
		private:
			N a[k+1];

			autodiff(N ra[k+1])
			{
				for ( size_t i = 0; i < k+1; ++i )
					a[i] = ra[i];
			}
			autodiff() {};
		public:
			autodiff(autodiff<N,k> const & o)
			{
				for ( size_t i = 0; i < k+1; ++i )
					a[i] = o.a[i];
			}
			static autodiff<N,k> constructVariable(
				N const & val)
			{
				N a[k+1];

				a[0] = val;

				if ( k > 0 )
					a[1] = N(1.0);

				for ( size_t i = 2; i < k+1; ++i )
					a[i] = 0;

				return autodiff<N,k>(a);
			}
			static autodiff<N,k> constructConstant(
				N const & val)
			{
				N a[k+1];

				a[0] = val;
				for ( size_t i = 1; i < k+1; ++i )
					a[i] = N(0.0);

				return autodiff<N,k>(a);
			}
			autodiff<N,k> & operator=(autodiff<N,k> const & o)
			{
				if ( this != &o )
				{
					for ( size_t i = 0; i < k+1; ++i )
						a[i] = o.a[i];
				}
				return *this;
			}

			autodiff<N,k> const & operator+() const
			{
				return *this;
			}
			autodiff<N,k> operator-() const
			{
				autodiff<N,k> o;
				for ( size_t i = 0; i < k+1; ++i )
					o.a[i] = -a[i];
				return o;
			}
			N const & get(size_t k0)
			{
				return a[k0];
			}
			autodiff<N,k> operator+=(autodiff<N,k> const & O)
			{
				*this = *this+O;
				return *this;
			}
			autodiff<N,k> operator-=(autodiff<N,k> const & O)
			{
				*this = *this-O;
				return *this;
			}
			autodiff<N,k> operator*=(autodiff<N,k> const & O)
			{
				*this = *this*O;
				return *this;
			}
			autodiff<N,k> operator/=(autodiff<N,k> const & O)
			{
				*this = *this/O;
				return *this;
			}

			friend autodiff<N,k> operator+<>(
				autodiff<N,k> const & A,
				autodiff<N,k> const & B
			);
			friend autodiff<N,k> operator-<>(
				autodiff<N,k> const & A,
				autodiff<N,k> const & B
			);
			friend autodiff<N,k> operator*<>(
				autodiff<N,k> const & A,
				autodiff<N,k> const & B
			);
			friend autodiff<N,k> operator/<>(
				autodiff<N,k> const & A,
				autodiff<N,k> const & B
			);
			friend std::ostream & operator<< <>(
				std::ostream & out,
				autodiff<N,k> const & obj);

			friend autodiff<N,k> sin<>(autodiff<N,k> const & A);
			friend autodiff<N,k> cos<>(autodiff<N,k> const & A);
			friend autodiff<N,k> tan<>(autodiff<N,k> const & A);
			friend autodiff<N,k> cot<>(autodiff<N,k> const & A);
			friend autodiff<N,k> asin<>(autodiff<N,k> const & A);
			friend autodiff<N,k> acos<>(autodiff<N,k> const & A);
			friend autodiff<N,k> atan<>(autodiff<N,k> const & A);
			friend autodiff<N,k> acot<>(autodiff<N,k> const & A);
			friend autodiff<N,k> sinh<>(autodiff<N,k> const & A);
			friend autodiff<N,k> cosh<>(autodiff<N,k> const & A);
			friend autodiff<N,k> tanh<>(autodiff<N,k> const & A);
			friend autodiff<N,k> coth<>(autodiff<N,k> const & A);
			friend autodiff<N,k> asinh<>(autodiff<N,k> const & A);
			friend autodiff<N,k> acosh<>(autodiff<N,k> const & A);
			friend autodiff<N,k> atanh<>(autodiff<N,k> const & A);
			friend autodiff<N,k> acoth<>(autodiff<N,k> const & A);
	};
	
	size_t binCoeff(size_t n, size_t k)
	{
		if ( n > (2*k) )
		{
			size_t o = 1;

			for ( size_t i = n; i > (n-k); --i )
				o *= i;

			size_t u = 1;

			for ( size_t i = k; i > 0; --i )
				u *= i;

			return o/u;

		}
		else // if ( n <= (2*k) )
		{
			size_t o = 1;

			for ( size_t i = n; i > k; --i )
				o *= i;

			size_t u = 1;

			for ( size_t i = n-k; i > 0; --i )
				u *= i;

			return o/u;
		}
	}

	template<typename N,size_t k> std::ostream & operator<<(
		std::ostream & out,
		autodiff<N,k> const & obj)
	{
		out << "autodiff(";

		for ( size_t i = 0; i < k+1; ++i )
			out << obj.a[i] << ((i+1<k+1)?",":"");

		out << ")";

		return out;
	}

	template<typename N,size_t k>
	autodiff<N,k> operator+(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		)
	{
		autodiff<N,k> r;

		for ( size_t i = 0; i < k+1; ++i )
			r.a[i] = A.a[i]+B.a[i];

		return r;
	}
	template<typename N,size_t k>
	autodiff<N,k> operator-(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		)
	{
		autodiff<N,k> r;

		for ( size_t i = 0; i < k+1; ++i )
			r.a[i] = A.a[i]-B.a[i];

		return r;
	}
	template<typename N,size_t k>
	autodiff<N,k> operator*(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		)
	{
		autodiff<N,k> r;

		for ( size_t M = 0; M < k+1; ++M )
		{
			r.a[M] = N(0);

			for ( size_t K = 0; K <= M; ++K )
				r.a[M] += N(binCoeff(M,K))*A.a[K]*B.a[M-K];
		}

		return r;
	}
	template<typename N,size_t k>
	autodiff<N,k> operator/(
		autodiff<N,k> const & A,
		autodiff<N,k> const & B
		)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = A.a[0]/B.a[0];
			return r;
		}
		if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = A.a[0]/B.a[0];
			r.a[1] = (A.a[1]-r.a[0]*B.a[1])/B.a[0];
			return r;
		}
		else
		{
			std::cerr << "autodiff division for k > 1 not implemented." << std::endl;
			std::terminate();
		}
	}

	template<typename N,size_t k>
	autodiff<N,k> sin(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = sin(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = sin(A.a[0]);
			r.a[1] = cos(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff sin() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> cos(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = cos(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = cos(A.a[0]);
			r.a[1] = -sin(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff cos() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> tan(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = tan(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = tan(A.a[0]);
			r.a[1] = (N(1.0)+r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff tan() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> cot(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = N(1.0)/tan(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = N(1.0)/tan(A.a[0]);
			r.a[1] = (N(1.0)+r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff cot() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> asin(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = asin(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = asin(A.a[0]);
			r.a[1] = (N(1.0)/sqrt(N(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff asin() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> acos(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = acos(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = acos(A.a[0]);
			r.a[1] = -(N(1.0)/sqrt(N(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acos() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> atan(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = atan(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = atan(A.a[0]);
			r.a[1] = (N(1.0)/(N(1.0)+(A.a[0]*A.a[0])))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff atan() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> acot(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = acot(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = atan(A.a[0]);
			r.a[1] = -(N(1.0)/(N(1.0)+(A.a[0]*A.a[0])))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acot() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> sinh(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = sinh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = sinh(A.a[0]);
			r.a[1] = cosh(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff sinh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> cosh(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = cosh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = cosh(A.a[0]);
			r.a[1] = sinh(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff cosh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> tanh(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = tanh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = tanh(A.a[0]);
			r.a[1] = (N(1.0)-r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff tanh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> coth(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = coth(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = coth(A.a[0]);
			r.a[1] = (N(1.0)-r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff tanh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> asinh(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = asinh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = asinh(A.a[0]);
			r.a[1] = (N(1.0)/sqrt(N(1.0)+A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff asinh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> acosh(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = acosh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = acosh(A.a[0]);
			r.a[1] = (N(1.0)/sqrt(A.a[0]*A.a[0]-N(1.0)))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acosh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> atanh(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = atanh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = atanh(A.a[0]);
			r.a[1] = (N(1.0)/(N(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff atanh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<typename N,size_t k>
	autodiff<N,k> acoth(autodiff<N,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<N,k> r;
			r.a[0] = (N(1.0)/N(2.0))*log((N(1.0)+A.a[0])/(N(1.0)-A.a[0]));
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<N,k> r;
			r.a[0] = (N(1.0)/N(2.0))*log((N(1.0)+A.a[0])/(N(1.0)-A.a[0]));
			r.a[1] = (N(1.0)/(N(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acoth() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}

	template<size_t k>
	autodiff<double,k> sin(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::sin(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::sin(A.a[0]);
			r.a[1] = std::cos(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff sin() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> cos(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::cos(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::cos(A.a[0]);
			r.a[1] = -std::sin(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff cos() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> tan(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::tan(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::tan(A.a[0]);
			r.a[1] = (double(1.0)+r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff tan() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> cot(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = double(1.0)/std::tan(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = double(1.0)/std::tan(A.a[0]);
			r.a[1] = (double(1.0)+r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff cot() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> asin(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::asin(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::asin(A.a[0]);
			r.a[1] = (double(1.0)/sqrt(double(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff asin() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> acos(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::acos(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::acos(A.a[0]);
			r.a[1] = -(double(1.0)/std::sqrt(double(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acos() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> atan(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::atan(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::atan(A.a[0]);
			r.a[1] = (double(1.0)/(double(1.0)+(A.a[0]*A.a[0])))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff atan() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> acot(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::acot(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::atan(A.a[0]);
			r.a[1] = -(double(1.0)/(double(1.0)+(A.a[0]*A.a[0])))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acot() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> sinh(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::sinh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::sinh(A.a[0]);
			r.a[1] = std::cosh(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff sinh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> cosh(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::cosh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::cosh(A.a[0]);
			r.a[1] = std::sinh(A.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff cosh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> tanh(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::tanh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::tanh(A.a[0]);
			r.a[1] = (double(1.0)-r.a[0]*r.a[0])*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff tanh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> asinh(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::asinh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::asinh(A.a[0]);
			r.a[1] = (double(1.0)/std::sqrt(double(1.0)+A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff asinh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> acosh(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::acosh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::acosh(A.a[0]);
			r.a[1] = (double(1.0)/std::sqrt(A.a[0]*A.a[0]-double(1.0)))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acosh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> atanh(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = std::atanh(A.a[0]);
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = std::atanh(A.a[0]);
			r.a[1] = (double(1.0)/(double(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff atanh() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
	template<size_t k>
	autodiff<double,k> acoth(autodiff<double,k> const & A)
	{
		if ( k == 0 )
		{
			autodiff<double,k> r;
			r.a[0] = (double(1.0)/double(2.0))*std::log((double(1.0)+A.a[0])/(double(1.0)-A.a[0]));
			return r;
		}
		else if ( k == 1 )
		{
			autodiff<double,k> r;
			r.a[0] = (double(1.0)/double(2.0))*std::log((double(1.0)+A.a[0])/(double(1.0)-A.a[0]));
			r.a[1] = (double(1.0)/(double(1.0)-A.a[0]*A.a[0]))*A.a[1];
			return r;
		}
		else
		{
			std::cerr << "autodiff acoth() for k>1 not implemented !" << std::endl;
			std::terminate();
		}
	}
};
#endif
