/*                                                                           
**  fi_lib++  --- A fast interval library (Version 2.0)                     
**                                                                  
**  Copyright (C) 2001:                                                        
**                                                     
**  Werner Hofschuster, Walter Kraemer                               
**  Wissenschaftliches Rechnen/Softwaretechnologie (WRSWT)  
**  Universitaet Wuppertal, Germany                                           
**  Michael Lerch, German Tischler, Juergen Wolff von Gudenberg       
**  Institut fuer Informatik                                         
**  Universitaet Wuerzburg, Germany                                           
** 
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/
#ifndef _FILIB_CONFIG_H
#define _FILIB_CONFIG_H

// -----------------------------------------------------------------------
// Specifiy the endianess of your machine. PC's usually are little endian
// and workstations big endian.
// -----------------------------------------------------------------------
#define	FILIB_LITTLE_ENDIAN	1234
#define	FILIB_BIG_ENDIAN	4321

// for PC's
#define FILIB_BYTE_ORDER FILIB_LITTLE_ENDIAN

// for workstations 
//#define FILIB_BYTE_ORDER FILIB_BIG_ENDIAN

// -----------------------------------------------------------------------
// Define PRED_SUCC_TABLES to make fi_lib++ to use table driven versions
// for pred and succ. In general these are a little bit faster.
// -----------------------------------------------------------------------
/* #define FILIB_PRED_SUCC_TABLES */

// -----------------------------------------------------------------------
// Define FILIB_EXTENDED to use the extended interval system ("Simple" 
// system). Otherwise error messages will be launched in indeterminate
// cases.
// In the extended system the special intervals [ EMPTY ], [ ENTIRE ],
// [ -INF ] and [ +INF ] are provided.
// -----------------------------------------------------------------------
/* #define FILIB_EXTENDED */

// -----------------------------------------------------------------------
// Define FILIB_NATIVE_ROUNDING to use native directed rounding instead
// of portable "rounding" via pred()/succ().
// Native rounding is provided only for a limited range of machines and
// compilers (see file include/rounding.h).
// ----------------------------------------------------------------------- 
#if !defined(__sparc)
#define FILIB_NATIVE_ROUNDING 
#endif

// -----------------------------------------------------------------------
// Define FILIB_ENSURE_ROUND_NEAR to ensure rounding to nearest after
// each interval operation. This makes sense only if native rounding is
// chosen.
// -----------------------------------------------------------------------
//#define FILIB_ENSURE_ROUND_NEAR

// -----------------------------------------------------------------------
// Define FILIB_INLINE_ARITH if interval arithmetic operators should be 
// expanded inline.
// -----------------------------------------------------------------------
#define FILIB_INLINE_ARITH

// -----------------------------------------------------------------------
// Define FILIB_INLINE_ALL if all interval functions should be expanded 
// inline, including arithmetic operations.
// -----------------------------------------------------------------------
#define FILIB_INLINE_ALL

// -----------------------------------------------------------------------
// Define FILIB_INTERVAL_BY_VALUE if interval arguments should be passed 
// to methods and functions of the library by value. 
// If FILIB_INTERVAL_BY_VALUE is not defined, intervals arguments will be
// passed by const &.  
// -----------------------------------------------------------------------
#define FILIB_INTERVAL_BY_VALUE

// -----------------------------------------------------------------------
// Define FILIB_USE_LIBI77_IO if the runtime library of the interval GNU
// Fortan compiler should be used for interval I/O.
// -----------------------------------------------------------------------
//#define FILIB_USE_LIBI77_IO


// -----------------------------------------------------------------------
// Define FILIB_STD_HEADERS if the compiler uses the C++ standard header 
// file names, e.g. <iostream> instead of <iostream.h>. 
// -----------------------------------------------------------------------
#define FILIB_STD_HEADERS


// -----------------------------------------------------------------------
// Define FILIB_NAMESPACES if all interval stuff should be placed in 
// namespace filib.
// -----------------------------------------------------------------------
// #define FILIB_NAMESPACES


#endif














